"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (Object.hasOwnProperty.call(mod, k)) result[k] = mod[k];
    result["default"] = mod;
    return result;
};
Object.defineProperty(exports, "__esModule", { value: true });
const web3_1 = __importDefault(require("web3"));
const dc_configs_1 = require("dc-configs");
const dc_logging_1 = require("dc-logging");
const account_js_1 = require("eth-lib/lib/account.js");
const Utils = __importStar(require("./utils"));
const logger = new dc_logging_1.Logger("EthInstance");
class Eth {
    constructor(params) {
        this._params = params;
        this._sign = account_js_1.sign;
        this._recover = account_js_1.recover;
        this._web3 = new web3_1.default(new web3_1.default.providers.HttpProvider(params.httpProviderUrl));
        this._cache = { lastBalances: { bet: {}, eth: {} } };
        // Init ERC20 contract
        this._ERC20Contract = this.initContract(this._params.ERC20ContractInfo.abi, this._params.ERC20ContractInfo.address);
    }
    getAccount() {
        return this._account;
    }
    initContract(abi, address) {
        return new this._web3.eth.Contract(abi, address);
    }
    initAccount(privateKey) {
        if (!privateKey) {
            const errorMessage = typeof window === "undefined"
                ? `ENV variable ACCOUNT_PRIVATE_KEY required!
           set ENV variable ACCOUNT_PRIVATE_KEY and init again`
                : `Private key is undefined
           Please set private key in params and init again`;
            switch (process.env.DC_NETWORK) {
                case "ropsten":
                    logger.warn(`
            You can get account with test ETH and BETs , from our faucet https://faucet.dao.casino/ 
            or use this random ${this._web3.eth.accounts.create().privateKey},
            but send Ropsten ETH and BETs to it before using
          `);
                    break;
                case "sdk":
                    logger.warn(`
            For local SDK env you can use this privkey:
            0x8d5366123cb560bb606379f90a0bfd4769eecc0557f1b362dcae9012b548b1e5
          `);
                    break;
                default:
                    logger.warn(`
            You can use this privkey: ${this._web3.eth.accounts.create().privateKey},
            but be sure that account have ETH and BETs
          `);
                    break;
            }
            throw new Error(errorMessage);
        }
        this._account = this._web3.eth.accounts.privateKeyToAccount(privateKey);
    }
    /**
     *
     * @param privateKey
     * @param walletPassword use only in browser
     */
    saveWallet(privateKey, walletPassword) {
        if (typeof window !== "undefined" &&
            typeof walletPassword === "undefined") {
            throw new Error("walletPassword is not defined");
        }
        if (typeof privateKey === "undefined") {
            throw new Error("privateKey is not defined");
        }
        this._web3.eth.accounts.wallet.add(privateKey);
        if (walletPassword && typeof window !== "undefined") {
            this._web3.eth.accounts.wallet.save(walletPassword, this._params.walletName);
        }
    }
    loadWallet(walletPassword) {
        if (typeof walletPassword === "undefined") {
            throw new Error("walletPassword is not define");
        }
        this._web3.eth.accounts.wallet.load(walletPassword, this._params.walletName);
    }
    getWalletAccount() {
        return this._web3.eth.accounts.wallet[0];
    }
    signData(argsToSign) {
        const hash = Utils.sha3(...argsToSign);
        const privateKey = Utils.add0x(this._account.privateKey);
        return this._sign(hash, privateKey);
    }
    signHash(hash) {
        const privateKey = Utils.add0x(this._account.privateKey);
        return this._sign(hash, privateKey);
    }
    recover(hash, peerSign) {
        return this._recover(hash, peerSign);
    }
    getBlockNumber() {
        return this._web3.eth.getBlockNumber();
    }
    allowance(spender, address = this._account.address) {
        return this._ERC20Contract.methods
            .allowance(address, spender)
            .call()
            .then(weis => Utils.dec2bet(weis));
    }
    sendTransaction(contract, methodName, args) {
        return new Promise((resolve, reject) => {
            const from = this._account.address;
            const receipt = contract.methods[methodName](...args).send({
                from,
                gas: this._params.gasParams.limit,
                gasPrice: this._params.gasParams.price
            });
            logger.debug(`Sent transaction: 
        contract: ${contract.options.address}, 
        method: ${methodName},
        from: ${from},
        args: ${JSON.stringify(args)}
        gas: ${this._params.gasParams.limit}
        gasPrice: ${this._params.gasParams.price}
      `);
            // Repeat if error
            receipt.catch(err => {
                logger.error("_REPEAT sendTransaction: " + methodName, err);
                reject(err);
            });
            receipt.on("error", err => {
                logger.error("REPEAT sendTransaction: " + methodName, err);
                reject(err);
            });
            receipt.on("transactionHash", transactionHash => {
                logger.debug("TX hash", transactionHash);
            });
            receipt.on("confirmation", confirmationCount => {
                if (confirmationCount <= dc_configs_1.config.default.waitForConfirmations) {
                    logger.debug(`${methodName} confirmationCount: ${confirmationCount}`);
                }
                else {
                    const rcpt = receipt;
                    rcpt.off("confirmation");
                    logger.debug("Transaction success");
                    resolve({ status: "success" });
                }
            });
        });
    }
    async ERC20ApproveSafe(spender, amount, minAmount = amount) {
        const allowance = await this.allowance(spender);
        if (0 < allowance && allowance < minAmount) {
            await this.sendTransaction(this._ERC20Contract, "approve", [spender, 0]);
        }
        if (allowance < minAmount) {
            await this.sendTransaction(this._ERC20Contract, "approve", [
                spender,
                this._web3.utils.toWei(amount.toString())
            ]);
        }
        return allowance;
    }
    async getBalances(address = this._account.address) {
        try {
            const [bet, eth] = await Promise.all([
                this.getBetBalance(address),
                this.getEthBalance(address)
            ]);
            this._cache.lastBalances.bet = bet;
            this._cache.lastBalances.eth = eth;
            return this._cache.lastBalances;
        }
        catch (error) {
            throw error;
        }
    }
    async getEthBalance(address) {
        if (!address) {
            throw new Error("Empty address in ETH balance request");
        }
        try {
            const weiBalance = await this._web3.eth.getBalance(address);
            const bnBalance = this._web3.utils.fromWei(weiBalance, "ether");
            return {
                balance: Number(bnBalance),
                updated: Date.now()
            };
        }
        catch (error) {
            throw error;
        }
    }
    async getBetBalance(address) {
        if (!address) {
            throw new Error("Empty address in BET balance request");
        }
        try {
            const decBalance = await this._ERC20Contract.methods
                .balanceOf(address)
                .call();
            const balance = Utils.dec2bet(decBalance);
            return {
                balance,
                updated: Date.now()
            };
        }
        catch (error) {
            throw error;
        }
    }
}
exports.Eth = Eth;
//# sourceMappingURL=Eth.js.map